package core;

import java.util.*;
import java.io.*;
import javax.media.j3d.*;

import core.*;

public class Simulation {
  public static Hashtable sysModels = new Hashtable();
  public static Hashtable cellModels = new Hashtable();
  public static Hashtable parsers = new Hashtable();

  private Properties configuration = new Properties();
  private BufferedReader[] cellProgramReaders;
  private SimFileTokenizer simFileTokenizer;
  private BufferedWriter outputWriter;
  private BufferedWriter statsWriter;
  private BufferedWriter errorWriter;

  private String systemModelName;
  private String systemModelOptions;
  private String cellModelName;
  private String cellModelOptions;
  private String parserModelName;
  private String parserModelOptions;

  private SimToken nextSimToken;

  private SystemModel sys;
  private CellProgram[] cellPrograms;

  private int step = 0;

  public Simulation(Vector ioObjects) {
    simFileTokenizer = new SimFileTokenizer( (BufferedReader) ioObjects.
                                            elementAt(0));
    //cellProgramReader = (BufferedReader) ioObjects.elementAt(1);
    cellProgramReaders = new BufferedReader[2];

    try {
      //eventually should not be hard coded
      cellProgramReaders[0] = new BufferedReader(new FileReader(
          "C:\\CellSim\\blastula2.cpa"));

      cellProgramReaders[1] = new BufferedReader(new FileReader(
          "C:\\CellSim\\blastula3.cpa"));
    }
    catch (FileNotFoundException e) {
      System.err.println("Unable to open file: " + e.toString());
      return;
    }
    catch (IOException e) {
      System.err.println(e.toString());
      return;
    }

    outputWriter = (BufferedWriter) ioObjects.elementAt(2);
    statsWriter = (BufferedWriter) ioObjects.elementAt(3);
    errorWriter = (BufferedWriter) ioObjects.elementAt(4);

    try {
      boolean sysModelRead = false, cellModelRead = false, parserModelRead = false;

      nextSimToken = simFileTokenizer.nextSimToken();
      while ( (nextSimToken.getType() == SimToken.SIM_CONFIG)
             && (!sysModelRead || !cellModelRead || !parserModelRead)) {
        if (nextSimToken.getType() == SimToken.SIM_CONFIG) {
          if (nextSimToken.getConfigParam().equals("system")) {
            configuration.setProperty("system", nextSimToken.getConfigValue());
            sysModelRead = true;
          }
          else if (nextSimToken.getConfigParam().equals("cell")) {
            configuration.setProperty("cell", nextSimToken.getConfigValue());
            cellModelRead = true;
          }
          else if (nextSimToken.getConfigParam().equals("parser")) {
            configuration.setProperty("parser", nextSimToken.getConfigValue());
            parserModelRead = true;
          }
        }
        nextSimToken = simFileTokenizer.nextSimToken();
      }

      if (! (sysModelRead && cellModelRead && parserModelRead)) {
        errorWriter.write("Incomplete specifications in SIM file");
        if (!sysModelRead) {
          errorWriter.write("System Model not specified");
        }
        else if (!cellModelRead) {
          errorWriter.write("Cell Model not specified");
        }
        else {
          errorWriter.write("Parser Model not specified");
        }
      }
    }
    catch (IOException e) {
      System.err.println("Io Exception ocurred : " + e.toString());
      System.exit(0);
    }
    systemModelName = new String(configuration.getProperty("system"));
    systemModelOptions = new String("");
    cellModelName = new String(configuration.getProperty("cell"));
    cellModelOptions = new String("");
    parserModelName = new String(configuration.getProperty("parser"));
    parserModelOptions = new String("");

    try {
      cellPrograms = createCellPrograms(parserModelName, parserModelOptions,
                                        cellProgramReaders);
      sys = createSystem(systemModelName, systemModelOptions);
    }
    catch (Exception e) {
      System.err.println(e.toString());
      System.exit(0);
    }
    runInstructions();
  }

  public BranchGroup getDisplayGraph() {
    return sys.getDisplayGraph();
  }

  public Hashtable getStatistics() {
    Vector argVector = new Vector();
    argVector.add(new String("Get Statistics"));
    Vector args = sys.provideService(argVector);

    /* Filter out and return the requested statistics */
    return (Hashtable) args.elementAt(0);
  }

  public CellProgram[] getCellPrograms() {
    return cellPrograms;
  }

  private void runInstructions() {
    while (nextSimToken.getType() == nextSimToken.SIM_INSTRUCTION
           && nextSimToken.getInstructionNumber() <= getStep()) {
      Vector args = new Vector();
      args.add(new String("Simulation Instruction"));
      args.add(nextSimToken.getInstruction());
      sys.provideService(args);
      try {
        nextSimToken = simFileTokenizer.nextSimToken();
      }
      catch (IOException e) {
        System.err.println(e.toString());
      }
    }
  }

  public void nextStep() {
    step++;
    runInstructions();
    sys.runCellProgramForOneStep();
  }

  public int getStep() {
    return step;
  }

  public void writeOutputLine(String s) throws IOException {
    outputWriter.write(s, 0, s.length());
    outputWriter.newLine();
  }

  public void writeStatsLine(String s) throws IOException {
    statsWriter.write(s, 0, s.length());
    statsWriter.newLine();
  }

  public void writeErrorLine(String s) throws IOException {
    errorWriter.write(s, 0, s.length());
    errorWriter.newLine();
  }

  public void finish() {
    try {
      outputWriter.close();
      statsWriter.close();
      errorWriter.close();
    }
    catch (IOException e) {
    }
  }

  public Cell createCell(Integer p_ID, SystemModel p_sys, CellProgram[] p_cps,
                         Vector args) throws Exception {
    CellFactory s = (CellFactory) cellModels.get(cellModelName);
    if (s == null) { // shape not found
      try {
        Class.forName(cellModelName);
        // By now the static block with no function would
        // have been executed if the system was found.
        // the system is expected to have put its factory
        // in the hashtable.
        s = (CellFactory) cellModels.get(cellModelName);
        if (s == null) { // if the system factory is not there even now
          throw (new Exception("System not created"));
        }
      }
      catch (ClassNotFoundException e) {
        // We'll throw an exception to indicate that
        // the shape could not be created
        throw (new Exception("System not created"));
      }
    }
    return s.create(p_ID, p_sys, p_cps, cellModelOptions, args);
  }

  public CellProgram createCellProgram(String modelName, String modelOptions,
                                       BufferedReader inp) throws
      GenericParseException, Exception {
    ParserFactory s = (ParserFactory) parsers.get(modelName);
    if (s == null) { // shape not found
      try {
        Class.forName(modelName);
        // By now the static block with no function would
        // have been executed if the Parser was found.
        // the Parser is expected to have put its factory
        // in the hashtable.
        s = (ParserFactory) parsers.get(modelName);
        if (s == null) { // if the system factory is not there even now
          throw (new Exception("Parser not created"));
        }
      }
      catch (ClassNotFoundException e) {
        // We'll throw an exception to indicate that
        // the shape could not be created
        throw (new Exception("Parser not created"));
      }
    }
    Parser p = s.create(inp, modelOptions);
    CellProgram cp;
    try {
      cp = p.getCellProgram();
    }
    catch (core.GenericParseException e) {
      throw new GenericParseException(e.toString());
    }
    return cp;
  }

  public CellProgram[] createCellPrograms(String modelName,
                                          String modelOptions,
                                          BufferedReader[] inp) throws
      GenericParseException, Exception {
    int numCellPrograms = inp.length;
    CellProgram[] cellPrograms = new CellProgram[numCellPrograms];
    for (int cellProgram = 0; cellProgram < numCellPrograms; cellProgram++) {
      cellPrograms[cellProgram] = createCellProgram(modelName,
          modelOptions,
          inp[cellProgram]);
    }
    return cellPrograms;
  }

  public SystemModel createSystem(String modelName, String modelOptions) throws
      Exception {
    SystemFactory s = (SystemFactory) sysModels.get(modelName);
    if (s == null) { // shape not found
      try {
        Class.forName(modelName);
        // By now the static block with no function would
        // have been executed if the System was found.
        // the System is expected to have put its factory
        // in the hashtable.
        s = (SystemFactory) sysModels.get(modelName);
        if (s == null) { // if the system factory is not there even now
          throw (new Exception("System not created"));
        }
      }
      catch (ClassNotFoundException e) {
        // We'll throw an exception to indicate that
        // the shape could not be created
        throw (new Exception("System not created"));
      }
      catch (Exception e) {
        throw (new Exception("Unknown Exception occurred"));
      }
    }
    return s.create(this, modelOptions);
  }
}
